@@ -11,6 +11,7 @@ import threading
 import shutil
 import functools
 import humanize
+import datetime as dt
 from time import time, sleep
 from codecs import open
 
@@ -52,9 +53,9 @@ from .__init__ import __version__
 click.disable_unicode_literals_warning = True
 
 try:
-    from urlparse import urlparse, unquote
+    from urlparse import urlparse, unquote, parse_qs
 except ImportError:
-    from urllib.parse import urlparse, unquote
+    from urllib.parse import urlparse, unquote, parse_qs
 
 from getpass import getuser
 from psycopg2 import OperationalError
@@ -75,6 +76,13 @@ MetaQuery = namedtuple(
     ])
 MetaQuery.__new__.__defaults__ = ('', False, 0, False, False, False, False)
 
+OutputSettings = namedtuple(
+    'OutputSettings',
+    'table_format dcmlfmt floatfmt missingval expanded max_width case_function'
+)
+OutputSettings.__new__.__defaults__ = (
+    None, None, None, '<null>', False, None, lambda x: x
+)
 
 # no-op logging handler
 class NullHandler(logging.Handler):
@@ -85,6 +93,7 @@ class NullHandler(logging.Handler):
 class PGCli(object):
 
     default_prompt = '\\u@\\h:\\d> '
+    max_len_prompt = 30
 
     def set_default_pager(self, config):
         configured_pager = config['main'].get('pager')
@@ -104,7 +113,7 @@ class PGCli(object):
 
     def __init__(self, force_passwd_prompt=False, never_passwd_prompt=False,
                  pgexecute=None, pgclirc_file=None, row_limit=None,
-                 single_connection=False):
+                 single_connection=False, less_chatty=None, prompt=None):
 
         self.force_passwd_prompt = force_passwd_prompt
         self.never_passwd_prompt = never_passwd_prompt
@@ -123,22 +132,29 @@ class PGCli(object):
         self.multi_line = c['main'].as_bool('multi_line')
         self.multiline_mode = c['main'].get('multi_line_mode', 'psql')
         self.vi_mode = c['main'].as_bool('vi')
+        self.auto_expand = c['main'].as_bool('auto_expand')
+        self.expanded_output = c['main'].as_bool('expand')
         self.pgspecial.timing_enabled = c['main'].as_bool('timing')
         if row_limit is not None:
             self.row_limit = row_limit
         else:
             self.row_limit = c['main'].as_int('row_limit')
 
+        self.min_num_menu_lines = c['main'].as_int('min_num_menu_lines')
+        self.multiline_continuation_char = c['main']['multiline_continuation_char']
         self.table_format = c['main']['table_format']
         self.syntax_style = c['main']['syntax_style']
         self.cli_style = c['colors']
         self.wider_completion_menu = c['main'].as_bool('wider_completion_menu')
-        self.less_chatty = c['main'].as_bool('less_chatty')
+        self.less_chatty = bool(less_chatty) or c['main'].as_bool('less_chatty')
         self.null_string = c['main'].get('null_string', '<null>')
-        self.prompt_format = c['main'].get('prompt', self.default_prompt)
+        self.prompt_format = prompt if prompt is not None else c['main'].get('prompt', self.default_prompt)
         self.on_error = c['main']['on_error'].upper()
         self.decimal_format = c['data_formats']['decimal']
         self.float_format = c['data_formats']['float']
+
+        self.now = dt.datetime.today()
+
         self.completion_refresher = CompletionRefresher()
 
         self.query_history = []
@@ -151,7 +167,11 @@ class PGCli(object):
             'generate_casing_file': c['main'].as_bool('generate_casing_file'),
             'generate_aliases': c['main'].as_bool('generate_aliases'),
             'asterisk_column_order': c['main']['asterisk_column_order'],
+            'qualify_columns': c['main']['qualify_columns'],
+            'case_column_headers': c['main'].as_bool('case_column_headers'),
+            'search_path_filter': c['main'].as_bool('search_path_filter'),
             'single_connection': single_connection,
+            'less_chatty': less_chatty,
             'keyword_casing': keyword_casing,
         }
 
@@ -272,13 +292,26 @@ class PGCli(object):
     def connect_uri(self, uri):
         uri = urlparse(uri)
         database = uri.path[1:]  # ignore the leading fwd slash
-        arguments = [database, uri.hostname, uri.username,
-                     uri.port, uri.password]
-        # unquote each URI part (they may be percent encoded)
-        self.connect(*list(map(lambda p: unquote(str(p)) if p else p, arguments)))
+
+        def fixup_possible_percent_encoding(s):
+            return unquote(str(s)) if s else s
+
+        arguments = dict(database=fixup_possible_percent_encoding(database),
+                         host=fixup_possible_percent_encoding(uri.hostname),
+                         user=fixup_possible_percent_encoding(uri.username),
+                         port=fixup_possible_percent_encoding(uri.port),
+                         passwd=fixup_possible_percent_encoding(uri.password))
+        # Deal with extra params e.g. ?sslmode=verify-ca&ssl-cert=/mycert
+        if uri.query:
+            arguments = dict(
+                {k: v for k, (v,) in parse_qs(uri.query).items()},
+                **arguments)
+
+        # unquote str(each URI part (they may be percent encoded)
+        self.connect(**arguments)
 
     def connect(self, database='', host='', user='', port='', passwd='',
-                dsn=''):
+                dsn='', **kwargs):
         # Connect to the database.
 
         if not user:
@@ -311,14 +344,15 @@ class PGCli(object):
         # a password (no -w flag), prompt for a passwd and try again.
         try:
             try:
-                pgexecute = PGExecute(database, user, passwd, host, port, dsn)
+                pgexecute = PGExecute(database, user, passwd, host, port, dsn,
+                                      **kwargs)
             except OperationalError as e:
                 if ('no password supplied' in utf8tounicode(e.args[0]) and
                         auto_passwd_prompt):
                     passwd = click.prompt('Password', hide_input=True,
                                           show_default=False, type=str)
                     pgexecute = PGExecute(database, user, passwd, host, port,
-                                          dsn)
+                                          dsn, **kwargs)
                 else:
                     raise e
 
@@ -343,16 +377,23 @@ class PGCli(object):
         :param document: Document
         :return: Document
         """
+        # FIXME: using application.pre_run_callables like this here is not the best solution.
+        # It's internal api of prompt_toolkit that may change. This was added to fix #668.
+        # We may find a better way to do it in the future.
+        saved_callables = cli.application.pre_run_callables
         while special.editor_command(document.text):
             filename = special.get_filename(document.text)
-            sql, message = special.open_external_editor(filename,
-                                                          sql=document.text)
+            query = (special.get_editor_query(document.text) or
+                     self.get_last_query())
+            sql, message = special.open_external_editor(filename, sql=query)
             if message:
                 # Something went wrong. Raise an exception and bail.
                 raise RuntimeError(message)
             cli.current_buffer.document = Document(sql, cursor_position=len(sql))
-            document = cli.run(False)
+            cli.application.pre_run_callables = []
+            document = cli.run()
             continue
+        cli.application.pre_run_callables = saved_callables
         return document
 
     def execute_command(self, text, query):
@@ -368,13 +409,9 @@ class PGCli(object):
         except NotImplementedError:
             click.secho('Not Yet Implemented.', fg="yellow")
         except OperationalError as e:
-            if ('server closed the connection'
-                    in utf8tounicode(e.args[0])):
-                self._handle_server_closed_connection()
-            else:
-                logger.error("sql: %r, error: %r", text, e)
-                logger.error("traceback: %r", traceback.format_exc())
-                click.secho(str(e), err=True, fg='red')
+            logger.error("sql: %r, error: %r", text, e)
+            logger.error("traceback: %r", traceback.format_exc())
+            self._handle_server_closed_connection()
         except Exception as e:
             logger.error("sql: %r, error: %r", text, e)
             logger.error("traceback: %r", traceback.format_exc())
@@ -439,7 +476,7 @@ class PGCli(object):
 
         try:
             while True:
-                document = self.cli.run(True)
+                document = self.cli.run()
 
                 # The reason we check here instead of inside the pgexecute is
                 # because we want to raise the Exit exception which will be
@@ -471,6 +508,8 @@ class PGCli(object):
                 else:
                     query = self.execute_command(document.text, query)
 
+                self.now = dt.datetime.today()
+
                 # Allow PGCompleter to learn user's preferred keywords, etc.
                 with self._completer_lock:
                     self.completer.extend_query_history(document.text)
@@ -491,17 +530,24 @@ class PGCli(object):
             set_vi_mode_enabled=set_vi_mode)
 
         def prompt_tokens(_):
-            return [(Token.Prompt, self.get_prompt(self.prompt_format))]
+            prompt = self.get_prompt(self.prompt_format)
+            if (self.prompt_format == self.default_prompt and
+               len(prompt) > self.max_len_prompt):
+                prompt = self.get_prompt('\\d> ')
+            return [(Token.Prompt, prompt)]
 
         def get_continuation_tokens(cli, width):
-            return [(Token.Continuation, '.' * (width - 1) + ' ')]
+            continuation=self.multiline_continuation_char * (width - 1) + ' '
+            return [(Token.Continuation, continuation)]
 
         get_toolbar_tokens = create_toolbar_tokens_func(
-            lambda: self.vi_mode, self.completion_refresher.is_refreshing)
+            lambda: self.vi_mode, self.completion_refresher.is_refreshing,
+            self.pgexecute.failed_transaction,
+            self.pgexecute.valid_transaction)
 
         layout = create_prompt_layout(
             lexer=PygmentsLexer(PostgresLexer),
-            reserve_space_for_menu=4,
+            reserve_space_for_menu=self.min_num_menu_lines,
             get_prompt_tokens=prompt_tokens,
             get_continuation_tokens=get_continuation_tokens,
             get_bottom_toolbar_tokens=get_toolbar_tokens,
@@ -581,14 +627,25 @@ class PGCli(object):
                     click.secho("Aborted!", err=True, fg='red')
                     break
 
-            if self.pgspecial.auto_expand:
+            if self.pgspecial.auto_expand or self.auto_expand:
                 max_width = self.cli.output.get_size().columns
             else:
                 max_width = None
 
-            formatted = format_output(
-                title, cur, headers, status, self.table_format, self.decimal_format,
-                self.float_format, self.null_string,  self.pgspecial.expanded_output, max_width)
+            expanded = self.pgspecial.expanded_output or self.expanded_output
+            settings = OutputSettings(
+                table_format=self.table_format,
+                dcmlfmt=self.decimal_format,
+                floatfmt=self.float_format,
+                missingval=self.null_string,
+                expanded=expanded,
+                max_width=max_width,
+                case_function=(
+                    self.completer.case if self.settings['case_column_headers']
+                    else lambda x: x
+                )
+            )
+            formatted = format_output(title, cur, headers, status, settings)
 
             output.extend(formatted)
             total = time() - start
@@ -686,6 +743,7 @@ class PGCli(object):
                 Document(text=text, cursor_position=cursor_positition), None)
 
     def get_prompt(self, string):
+        string = string.replace('\\t', self.now.strftime('%x %X'))
         string = string.replace('\\u', self.pgexecute.user or '(none)')
         string = string.replace('\\h', self.pgexecute.host or '(none)')
         string = string.replace('\\d', self.pgexecute.dbname or '(none)')
@@ -695,6 +753,11 @@ class PGCli(object):
         string = string.replace('\\n', "\n")
         return string
 
+    def get_last_query(self):
+        """Get the last query executed or None."""
+        return self.query_history[-1][0] if self.query_history else None
+
+
 
 @click.command()
 # Default host is '' so psycopg2 can default to either localhost or unix socket
@@ -702,8 +765,8 @@ class PGCli(object):
         help='Host address of the postgres database.')
 @click.option('-p', '--port', default=5432, help='Port number at which the '
         'postgres instance is listening.', envvar='PGPORT')
-@click.option('-U', '--user', envvar='PGUSER', help='User name to '
-        'connect to the postgres database.')
+@click.option('-U', '--username', 'username_opt', envvar='PGUSER',
+        help='Username to connect to the postgres database.')
 @click.option('-W', '--password', 'prompt_passwd', is_flag=True, default=False,
         help='Force password prompt.')
 @click.option('-w', '--no-password', 'never_prompt', is_flag=True,
@@ -718,12 +781,17 @@ class PGCli(object):
         envvar='PGCLIRC', help='Location of pgclirc file.')
 @click.option('-D', '--dsn', default='', envvar='DSN',
         help='Use DSN configured into the [alias_dsn] section of pgclirc file.')
-@click.option('-R', '--row-limit', default=None, envvar='PGROWLIMIT', type=click.INT,
+@click.option('--row-limit', default=None, envvar='PGROWLIMIT', type=click.INT,
         help='Set threshold for row limit prompt. Use 0 to disable prompt.')
+@click.option('--less-chatty', 'less_chatty', is_flag=True,
+        default=False,
+        help='Skip intro on startup and goodbye on exit.')
+@click.option('--prompt', help='Prompt format (Default: "\\u@\\h:\\d> ").')
 @click.argument('database', default=lambda: None, envvar='PGDATABASE', nargs=1)
 @click.argument('username', default=lambda: None, envvar='PGUSER', nargs=1)
-def cli(database, user, host, port, prompt_passwd, never_prompt,
-    single_connection, dbname, username, version, pgclirc, dsn, row_limit):
+def cli(database, username_opt, host, port, prompt_passwd, never_prompt,
+    single_connection, dbname, username, version, pgclirc, dsn, row_limit,
+    less_chatty, prompt):
 
     if version:
         print('Version:', __version__)
@@ -746,11 +814,12 @@ def cli(database, user, host, port, prompt_passwd, never_prompt,
                    config_full_path)
 
     pgcli = PGCli(prompt_passwd, never_prompt, pgclirc_file=pgclirc,
-                  row_limit=row_limit, single_connection=single_connection)
+                  row_limit=row_limit, single_connection=single_connection,
+                  less_chatty=less_chatty, prompt=prompt)
 
     # Choose which ever one has a valid value.
     database = database or dbname
-    user = username or user
+    user = username_opt or username
 
     if dsn is not '':
         try:
@@ -793,13 +862,19 @@ def obfuscate_process_password():
     setproctitle.setproctitle(process_title)
 
 
-def format_output(title, cur, headers, status, table_format, dcmlfmt, floatfmt,
-                  missingval='<null>', expanded=False, max_width=None):
+def format_output(title, cur, headers, status, settings):
     output = []
+    missingval = settings.missingval
+    table_format = settings.table_format
+    dcmlfmt = settings.dcmlfmt
+    floatfmt = settings.floatfmt
+    expanded = settings.expanded
+    max_width = settings.max_width
+    case_function = settings.case_function
     if title:  # Only print the title if it's not None.
         output.append(title)
     if cur:
-        headers = [utf8tounicode(x) for x in headers]
+        headers = [case_function(utf8tounicode(x)) for x in headers]
         if expanded and headers:
             output.append(expanded_table(cur, headers, missingval))
         else:
