@@ -41,10 +41,13 @@ View.__new__.__defaults__ = (None, tuple())
 FromClauseItem.__new__.__defaults__ = (None, tuple(), tuple())
 
 Column = namedtuple(
-    'Column', ['table_refs', 'require_last_table', 'local_tables'])
-Column.__new__.__defaults__ = (None, None, tuple())
+    'Column',
+    ['table_refs', 'require_last_table', 'local_tables', 'qualifiable']
+)
+Column.__new__.__defaults__ = (None, None, tuple(), False)
 
-Keyword = namedtuple('Keyword', [])
+Keyword = namedtuple('Keyword', ['last_token'])
+Keyword.__new__.__defaults__ = (None,)
 NamedQuery = namedtuple('NamedQuery', [])
 Datatype = namedtuple('Datatype', ['schema'])
 Alias = namedtuple('Alias', ['aliases'])
@@ -166,6 +169,26 @@ def _strip_named_query(txt):
         txt = named_query_regex.sub('', txt)
     return txt
 
+function_body_pattern = re.compile('(\\$.*?\\$)([\s\S]*?)\\1', re.M)
+
+
+def _find_function_body(text):
+    split = function_body_pattern.search(text)
+    return (split.start(2), split.end(2)) if split else (None, None)
+
+
+def _statement_from_function(full_text, text_before_cursor, statement):
+    current_pos = len(text_before_cursor)
+    body_start, body_end = _find_function_body(full_text)
+    if body_start is None:
+        return full_text, text_before_cursor, statement
+    if not body_start <= current_pos < body_end:
+        return full_text, text_before_cursor, statement
+    full_text = full_text[body_start:body_end]
+    text_before_cursor = text_before_cursor[body_start:]
+    parsed = sqlparse.parse(text_before_cursor)
+    return _split_multiple_statements(full_text, text_before_cursor, parsed)
+
 
 def _split_multiple_statements(full_text, text_before_cursor, parsed):
     if len(parsed) > 1:
@@ -189,11 +212,29 @@ def _split_multiple_statements(full_text, text_before_cursor, parsed):
         statement = parsed[0]
     else:
         # The empty string
-        statement = None
+        return full_text, text_before_cursor, None
 
+    token2 = None
+    if statement.get_type() in ('CREATE', 'CREATE OR REPLACE'):
+        token1 = statement.token_first()
+        if token1:
+            token1_idx = statement.token_index(token1)
+            token2 = statement.token_next(token1_idx)[1]
+    if token2 and token2.value.upper() == 'FUNCTION':
+        full_text, text_before_cursor, statement = _statement_from_function(
+            full_text, text_before_cursor, statement
+        )
     return full_text, text_before_cursor, statement
 
 
+SPECIALS_SUGGESTION = {
+    'dT': Datatype,
+    'df': Function,
+    'dt': Table,
+    'dv': View,
+    'sf': Function,
+}
+
 def suggest_special(text):
     text = text.lstrip()
     cmd, _, arg = parse_special_command(text)
@@ -221,7 +262,7 @@ def suggest_special(text):
         schema = None
 
     if cmd[1:] == 'd':
-        # \d can descibe tables or views
+        # \d can describe tables or views
         if schema:
             return (Table(schema=schema),
                     View(schema=schema),)
@@ -229,12 +270,8 @@ def suggest_special(text):
             return (Schema(),
                     Table(schema=None),
                     View(schema=None),)
-    elif cmd[1:] in ('dt', 'dv', 'df', 'dT'):
-        rel_type = {'dt': Table,
-                    'dv': View,
-                    'df': Function,
-                    'dT': Datatype,
-                    }[cmd[1:]]
+    elif cmd[1:] in SPECIALS_SUGGESTION:
+        rel_type = SPECIALS_SUGGESTION[cmd[1:]]
         if schema:
             return (rel_type(schema=schema),)
         else:
@@ -340,11 +377,11 @@ def suggest_based_on_last_token(token, stmt):
             return (Column(table_refs=stmt.get_tables('insert')),)
         # We're probably in a function argument list
         return (Column(table_refs=extract_tables(stmt.full_text),
-                       local_tables=stmt.local_tables),)
-    elif token_v in ('set', 'by', 'distinct'):
+                       local_tables=stmt.local_tables, qualifiable=True),)
+    elif token_v == 'set':
         return (Column(table_refs=stmt.get_tables(),
                        local_tables=stmt.local_tables),)
-    elif token_v in ('select', 'where', 'having'):
+    elif token_v in ('select', 'where', 'having', 'by', 'distinct'):
         # Check for a table alias or schema qualification
         parent = (stmt.identifier and stmt.identifier.get_parent_name()) or []
         tables = stmt.get_tables()
@@ -355,9 +392,10 @@ def suggest_based_on_last_token(token, stmt):
                     View(schema=parent),
                     Function(schema=parent),)
         else:
-            return (Column(table_refs=tables, local_tables=stmt.local_tables),
+            return (Column(table_refs=tables, local_tables=stmt.local_tables,
+                           qualifiable=True),
                     Function(schema=None),
-                    Keyword(),)
+                    Keyword(token_v.upper()),)
     elif token_v == 'as':
         # Don't suggest anything for aliases
         return ()
@@ -454,6 +492,8 @@ def suggest_based_on_last_token(token, stmt):
         if not schema:
             suggestions.append(Schema())
         return tuple(suggestions)
+    elif token_v in {'alter', 'create', 'drop'}:
+        return (Keyword(token_v.upper()),)
     elif token.is_keyword:
         # token is a keyword we haven't implemented any special handling for
         # go backwards in the query until we find one we do recognize
@@ -461,7 +501,7 @@ def suggest_based_on_last_token(token, stmt):
         if prev_keyword:
             return suggest_based_on_last_token(prev_keyword, stmt)
         else:
-            return (Keyword(),)
+            return (Keyword(token_v.upper()),)
     else:
         return (Keyword(),)
 
