@@ -5,10 +5,10 @@
 from __future__ import print_function
 from __future__ import unicode_literals
 from collections import namedtuple
-from decimal import Decimal
 from platform import python_version_tuple
 from wcwidth import wcswidth
 from ..encodingutils import utf8tounicode
+from decimal import Decimal
 import re
 
 
@@ -302,15 +302,14 @@ def _isnumber(string):
 
 def _isint(string):
     """
-    >>> _isint("123")
+    >>> _isint(123)
     True
+    >>> _isint("123")
+    False
     >>> _isint("123.45")
     False
     """
-    return type(string) is _int_type or type(string) is _long_type or \
-           (isinstance(string, _binary_type) or isinstance(string, _text_type)) and \
-           _isconvertible(int, string)
-
+    return type(string) is _int_type or type(string) is _long_type
 
 def _type(string, has_invisible=True):
     """The least generic type (type(None), int, float, str, unicode).
@@ -320,7 +319,7 @@ def _type(string, has_invisible=True):
     >>> _type("foo") is type("")
     True
     >>> _type("1") is type(1)
-    True
+    False
     >>> _type('\x1b[31m42\x1b[0m') is type(42)
     True
     >>> _type('\x1b[31m42\x1b[0m') is type(42)
@@ -334,13 +333,13 @@ def _type(string, has_invisible=True):
 
     if string is None:
         return _none_type
-    if isinstance(string, (bool, Decimal,)):
+    if isinstance(string, bool):
         return _text_type
     elif hasattr(string, "isoformat"):  # datetime.datetime, date, and time
         return _text_type
     elif _isint(string):
         return int
-    elif _isnumber(string):
+    elif isinstance(string, (float, Decimal)):
         return float
     elif isinstance(string, _binary_type):
         return _binary_type
@@ -497,7 +496,7 @@ def _column_type(strings, has_invisible=True):
     return reduce(_more_generic, types, int)
 
 
-def _format(val, valtype, floatfmt, missingval=""):
+def _format(val, valtype, dcmlfmt, floatfmt, missingval=""):
     """Format a value accoding to its type.
 
     Unicode is supported:
@@ -512,7 +511,9 @@ def _format(val, valtype, floatfmt, missingval=""):
     if val is None:
         return missingval
 
-    if valtype in [int, _text_type]:
+    if valtype is int:
+        return format(int(val), dcmlfmt)
+    elif valtype is _text_type:
         return "{0}".format(val)
     elif valtype is _binary_type:
         try:
@@ -648,7 +649,7 @@ def _normalize_tabular_data(tabular_data, headers):
 
 
 def tabulate(tabular_data, headers=[], tablefmt="simple",
-             floatfmt="g", numalign="decimal", stralign="left",
+             dcmlfmt="d", floatfmt="g", numalign="decimal", stralign="left",
              missingval=""):
     """Format a fixed width table for pretty printing.
 
@@ -889,7 +890,8 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     _text_type_encode = lambda x: _text_type(utf8tounicode(x))
     plain_text = '\n'.join(['\t'.join(map(_text_type_encode, headers))] + \
                             ['\t'.join(map(_text_type_encode, row)) for row in list_of_lists])
-    has_invisible = re.search(_invisible_codes, plain_text)
+    has_invisible = (re.search(_invisible_codes, plain_text) or
+      re.search(_invisible_codes, missingval))
     if has_invisible:
         width_fn = _visible_width
     else:
@@ -898,7 +900,7 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     # format rows and columns, convert numeric values to strings
     cols = list(zip(*list_of_lists))
     coltypes = list(map(_column_type, cols))
-    cols = [[_format(v, ct, floatfmt, missingval) for v in c]
+    cols = [[_format(v, ct, dcmlfmt, floatfmt, missingval) for v in c]
              for c,ct in zip(cols, coltypes)]
 
     # align columns
