@@ -28,23 +28,33 @@ _logger = logging.getLogger(__name__)
 NamedQueries.instance = NamedQueries.from_config(
     load_config(config_location() + 'config'))
 
-
 Match = namedtuple('Match', ['completion', 'priority'])
+_SchemaObject = namedtuple('SchemaObject', ['name', 'schema', 'function'])
+def SchemaObject(name, schema=None, function=False):
+    return _SchemaObject(name, schema, function)
+
+_Candidate = namedtuple(
+    'Candidate', ['completion', 'prio', 'meta', 'synonyms', 'prio2']
+)
+def Candidate(completion, prio=None, meta=None, synonyms=None, prio2=None):
+    return _Candidate(completion, prio, meta, synonyms or [completion], prio2)
 
 normalize_ref = lambda ref: ref if ref[0] == '"' else '"' + ref.lower() +  '"'
 
-def generate_alias(tbl, tbs):
+def generate_alias(tbl):
     """ Generate a table alias, consisting of all upper-case letters in
     the table name, or, if there are no upper-case letters, the first letter +
     all letters preceded by _
     param tbl - unescaped name of the table to alias
-    param tbls - set TableReference objects for tables already in query
     """
     return ''.join([l for l in tbl if l.isupper()] or
         [l for l, prev in zip(tbl,  '_' + tbl) if prev == '_' and l != '_'])
 
 class PGCompleter(Completer):
-    keywords = get_literals('keywords')
+    # keywords_tree: A dict mapping keywords to well known following keywords.
+    # e.g. 'CREATE': ['TABLE', 'USER', ...],
+    keywords_tree = get_literals('keywords', type_=dict)
+    keywords = tuple(set(chain(keywords_tree.keys(), *keywords_tree.values())))
     functions = get_literals('functions')
     datatypes = get_literals('datatypes')
 
@@ -54,9 +64,12 @@ class PGCompleter(Completer):
         self.pgspecial = pgspecial
         self.prioritizer = PrevalenceCounter()
         settings = settings or {}
+        self.search_path_filter = settings.get('search_path_filter')
         self.generate_aliases = settings.get('generate_aliases')
         self.casing_file = settings.get('casing_file')
         self.generate_casing_file = settings.get('generate_casing_file')
+        self.qualify_columns = settings.get(
+            'qualify_columns', 'if_more_than_one_table')
         self.asterisk_column_order = settings.get(
             'asterisk_column_order', 'table_order')
 
@@ -237,15 +250,15 @@ class PGCompleter(Completer):
                            'datatypes': {}}
         self.all_completions = set(self.keywords + self.functions)
 
-    def find_matches(self, text, collection, mode='fuzzy',
-                     meta=None, meta_collection=None,
-                     priority_collection = None):
+    def find_matches(self, text, collection, mode='fuzzy', meta=None):
         """Find completion matches for the given text.
 
         Given the user's input text and a collection of available
         completions, find completions matching the last word of the
         text.
 
+        `collection` can be either a list of strings or a list of Candidate
+        namedtuples.
         `mode` can be either 'fuzzy', or 'strict'
             'fuzzy': fuzzy matching, ties broken by name prevalance
             `keyword`: start only matching, ties broken by keyword prevalance
@@ -254,7 +267,8 @@ class PGCompleter(Completer):
         in the collection of available completions.
 
         """
-
+        if not collection:
+            return []
         prio_order = [
             'keyword', 'function', 'view', 'table', 'datatype', 'database',
             'schema', 'column', 'table alias', 'join', 'name join', 'fk join'
@@ -287,6 +301,12 @@ class PGCompleter(Completer):
             pat = re.compile('(%s)' % regex)
 
             def _match(item):
+                if item.lower()[:len(text) + 1] in (text, text + ' '):
+                    # Exact match of first word in suggestion
+                    # This is to get exact alias matches to the top
+                    # E.g. for input `e`, 'Entries E' should be on top
+                    # (before e.g. `EndUsers EU`)
+                    return float('Infinity'), -1
                 r = pat.search(self.unescape_name(item.lower()))
                 if r:
                     return -len(r.group()), -r.start()
@@ -300,21 +320,24 @@ class PGCompleter(Completer):
                     # fuzzy matches
                     return -float('Infinity'), -match_point
 
-        # Fallback to meta param if meta_collection param is None
-        meta_collection = meta_collection or repeat(meta)
-        # Fallback to 0 if priority_collection param is None
-        priority_collection = priority_collection or repeat(0)
-
-        collection = zip(collection, meta_collection, priority_collection)
-
         matches = []
+        for cand in collection:
+            if isinstance(cand, _Candidate):
+                item, prio, display_meta, synonyms, prio2 = cand
+                if display_meta is None:
+                    display_meta = meta
+                syn_matches = (_match(x) for x in synonyms)
+                # Nones need to be removed to avoid max() crashing in Python 3
+                syn_matches = [m for m in syn_matches if m]
+                sort_key = max(syn_matches) if syn_matches else None
+            else:
+                item, display_meta, prio, prio2 = cand, meta, 0, 0
+                sort_key = _match(cand)
 
-        for item, meta, prio in collection:
-            sort_key = _match(item)
             if sort_key:
-                if meta and len(meta) > 50:
+                if display_meta and len(display_meta) > 50:
                     # Truncate meta-text to 50 characters, if necessary
-                    meta = meta[:47] + u'...'
+                    display_meta = display_meta[:47] + u'...'
 
                 # Lexical order of items in the collection, used for
                 # tiebreaking items with the same match group length and start
@@ -330,10 +353,14 @@ class PGCompleter(Completer):
                     + tuple(c for c in item))
 
                 item = self.case(item)
-                priority = sort_key, type_priority, prio, priority_func(item), lexical_priority
+                priority = (
+                    sort_key, type_priority, prio, priority_func(item),
+                    prio2, lexical_priority
+                )
 
                 matches.append(Match(
-                    completion=Completion(item, -text_len, display_meta=meta),
+                    completion=Completion(item, -text_len,
+                    display_meta=display_meta),
                     priority=priority))
         return matches
 
@@ -375,12 +402,21 @@ class PGCompleter(Completer):
 
     def get_column_matches(self, suggestion, word_before_cursor):
         tables = suggestion.table_refs
+        do_qualify = suggestion.qualifiable and {'always': True, 'never': False,
+            'if_more_than_one_table': len(tables) > 1}[self.qualify_columns]
+        qualify = lambda col, tbl: (
+            (tbl + '.' + self.case(col)) if do_qualify else self.case(col))
         _logger.debug("Completion column scope: %r", tables)
         scoped_cols = self.populate_scoped_cols(tables, suggestion.local_tables)
 
         colit = scoped_cols.items
-        flat_cols = list(chain(*((c.name for c in cols)
-            for t, cols in colit())))
+        def make_cand(name, ref):
+            synonyms = (name, generate_alias(self.case(name)))
+            return Candidate(qualify(name, ref), 0, 'column', synonyms)
+        flat_cols = []
+        for t, cols in colit():
+            for c in cols:
+                flat_cols.append(make_cand(c.name, t.ref))
         if suggestion.require_last_table:
             # require_last_table is used for 'tb11 JOIN tbl2 USING (...' which should
             # suggest only columns that appear in the last table and one more
@@ -399,14 +435,10 @@ class PGCompleter(Completer):
                 # User typed x.*; replicate "x." for all columns except the
                 # first, which gets the original (as we only replace the "*"")
                 sep = ', ' + word_before_cursor[:-1]
-                collist = sep.join(self.case(c) for c in flat_cols)
-            elif len(scoped_cols) > 1:
-                # Multiple tables; qualify all columns
-                collist = ', '.join(t.ref + '.' + self.case(c.name)
-                    for t, cs in colit() for c in cs)
+                collist = sep.join(self.case(c.completion) for c in flat_cols)
             else:
-                # Plain columns
-                collist = ', '.join(self.case(c) for c in flat_cols)
+                collist = ', '.join(qualify(c.name, t.ref)
+                    for t, cs in colit() for c in cs)
 
             return [Match(completion=Completion(collist, -1,
                 display_meta='columns', display='*'), priority=(1,1,1))]
@@ -422,7 +454,7 @@ class PGCompleter(Completer):
         tbl = self.case(tbl)
         tbls = set(normalize_ref(t.ref) for t in tbls)
         if self.generate_aliases:
-            tbl = generate_alias(self.unescape_name(tbl), tbls)
+            tbl = generate_alias(self.unescape_name(tbl))
         if normalize_ref(tbl) not in tbls:
             return tbl
         elif tbl[0] == '"':
@@ -440,7 +472,7 @@ class PGCompleter(Completer):
         refs = set(normalize_ref(t.ref) for t in tbls)
         other_tbls = set((t.schema, t.name)
             for t in list(cols)[:-1])
-        joins, prios = [], []
+        joins = []
         # Iterate over FKs in existing tables to find potential joins
         fks = ((fk, rtbl, rcol) for rtbl, rcols in cols.items()
             for rcol in rcols for fk in rcol.foreignkeys)
@@ -460,18 +492,20 @@ class PGCompleter(Completer):
             else:
                 join = '{0} ON {0}.{1} = {2}.{3}'.format(
                     c(left.tbl), c(left.col), rtbl.ref, c(right.col))
+            alias = generate_alias(self.case(left.tbl))
+            synonyms = [join, '{0} ON {0}.{1} = {2}.{3}'.format(
+                alias, c(left.col), rtbl.ref, c(right.col))]
             # Schema-qualify if (1) new table in same schema as old, and old
             # is schema-qualified, or (2) new in other schema, except public
             if not suggestion.schema and (qualified[normalize_ref(rtbl.ref)]
                 and left.schema == right.schema
                 or left.schema not in(right.schema, 'public')):
                 join = left.schema + '.' + join
-            joins.append(join)
-            prios.append(ref_prio[normalize_ref(rtbl.ref)] * 2 + (
-                0 if (left.schema, left.tbl) in other_tbls else 1))
+            prio = ref_prio[normalize_ref(rtbl.ref)] * 2 + (
+                0 if (left.schema, left.tbl) in other_tbls else 1)
+            joins.append(Candidate(join, prio, 'join', synonyms=synonyms))
 
-        return self.find_matches(word_before_cursor, joins, meta='join',
-            priority_collection=prios)
+        return self.find_matches(word_before_cursor, joins, meta='join')
 
     def get_join_condition_matches(self, suggestion, word_before_cursor):
         col = namedtuple('col', 'schema tbl col')
@@ -482,16 +516,15 @@ class PGCompleter(Completer):
             ltbl, lcols = [(t, cs) for (t, cs) in tbls() if t.ref == lref][-1]
         except IndexError: # The user typed an incorrect table qualifier
             return []
-        conds, prios, found_conds = [], [], set()
+        conds, found_conds = [], set()
 
-        def add_cond(lcol, rcol, rref, prio):
+        def add_cond(lcol, rcol, rref, prio, meta):
             prefix = '' if suggestion.parent else ltbl.ref + '.'
             case = self.case
             cond = prefix + case(lcol) + ' = ' + rref + '.' + case(rcol)
             if cond not in found_conds:
                 found_conds.add(cond)
-                conds.append(cond)
-                prios.append(prio + ref_prio[rref])
+                conds.append(Candidate(cond, prio + ref_prio[rref], meta))
 
         def list_dict(pairs): # Turns [(a, b), (a, c)] into {a: [b, c]}
             d = defaultdict(list)
@@ -514,35 +547,30 @@ class PGCompleter(Completer):
             par = col(fk.parentschema, fk.parenttable, fk.parentcolumn)
             left, right = (child, par) if left == child else (par, child)
             for rtbl in coldict[right]:
-                add_cond(left.col, right.col, rtbl.ref, 2000)
-        matches = self.find_matches(word_before_cursor, conds,
-          meta='fk join', priority_collection=prios)
+                add_cond(left.col, right.col, rtbl.ref, 2000, 'fk join')
         # For name matching, use a {(colname, coltype): TableReference} dict
         coltyp = namedtuple('coltyp', 'name datatype')
         col_table = list_dict((coltyp(c.name, c.datatype), t) for t, c in cols)
-        conds, prios = [], []
         # Find all name-match join conditions
         for c in (coltyp(c.name, c.datatype) for c in lcols):
             for rtbl in (t for t in col_table[c] if t.ref != ltbl.ref):
-                add_cond(c.name, c.name, rtbl.ref, 1000
-                    if c.datatype in ('integer', 'bigint', 'smallint') else 0)
+                prio = 1000 if c.datatype in (
+                    'integer', 'bigint', 'smallint') else 0
+                add_cond(c.name, c.name, rtbl.ref, prio, 'name join')
 
-        return matches + self.find_matches(word_before_cursor, conds,
-          meta='name join', priority_collection=prios)
+        return self.find_matches(word_before_cursor, conds, meta='join')
 
     def get_function_matches(self, suggestion, word_before_cursor, alias=False):
+        def _cand(func, alias):
+            return self._make_cand(func, alias, suggestion)
         if suggestion.filter == 'for_from_clause':
             # Only suggest functions allowed in FROM clause
             filt = lambda f: not f.is_aggregate and not f.is_window
-            funcs = self.populate_functions(suggestion.schema, filt)
-            if alias:
-                funcs = [self.case(f) + '() ' + self.alias(f,
-                    suggestion.table_refs) for f in funcs]
-            else:
-                funcs = [self.case(f) + '()' for f in funcs]
+            funcs = [_cand(f, alias)
+                     for f in self.populate_functions(suggestion.schema, filt)]
         else:
-            funcs = [f + '()' for f in self.populate_schema_objects(
-                suggestion.schema, 'functions')]
+            fs = self.populate_schema_objects(suggestion.schema, 'functions')
+            funcs = [_cand(f, alias=False) for f in fs]
 
         # Function overloading means we way have multiple functions of the same
         # name at this point, so keep unique names only
@@ -580,20 +608,30 @@ class PGCompleter(Completer):
             + self.get_view_matches(v_sug, word_before_cursor, alias)
             + self.get_function_matches(f_sug, word_before_cursor, alias))
 
+    # Note: tbl is a SchemaObject
+    def _make_cand(self, tbl, do_alias, suggestion):
+        cased_tbl = self.case(tbl.name)
+        if do_alias:
+            alias = self.alias(cased_tbl, suggestion.table_refs)
+        synonyms = (cased_tbl, generate_alias(cased_tbl))
+        maybe_parens = '()' if tbl.function else ''
+        maybe_alias = (' ' + alias) if do_alias else ''
+        maybe_schema = (self.case(tbl.schema) + '.') if tbl.schema else ''
+        item = maybe_schema + cased_tbl + maybe_parens + maybe_alias
+        prio2 = 0 if tbl.schema else 1
+        return Candidate(item, synonyms=synonyms, prio2=prio2)
+
     def get_table_matches(self, suggestion, word_before_cursor, alias=False):
         tables = self.populate_schema_objects(suggestion.schema, 'tables')
-        tables.extend(tbl.name for tbl in suggestion.local_tables)
+        tables.extend(SchemaObject(tbl.name) for tbl in suggestion.local_tables)
 
         # Unless we're sure the user really wants them, don't suggest the
         # pg_catalog tables that are implicitly on the search path
         if not suggestion.schema and (
                 not word_before_cursor.startswith('pg_')):
-            tables = [t for t in tables if not t.startswith('pg_')]
-        if alias:
-            tables = [self.case(t) + ' ' + self.alias(t, suggestion.table_refs)
-                for t in tables]
-        return self.find_matches(word_before_cursor, tables,
-            meta='table')
+            tables = [t for t in tables if not t.name.startswith('pg_')]
+        tables = [self._make_cand(t, alias, suggestion) for t in tables]
+        return self.find_matches(word_before_cursor, tables, meta='table')
 
 
     def get_view_matches(self, suggestion, word_before_cursor, alias=False):
@@ -601,10 +639,8 @@ class PGCompleter(Completer):
 
         if not suggestion.schema and (
                 not word_before_cursor.startswith('pg_')):
-            views = [v for v in views if not v.startswith('pg_')]
-        if alias:
-            views = [self.case(v) + ' ' + self.alias(v, suggestion.table_refs)
-                for v in views]
+            views = [v for v in views if not v.name.startswith('pg_')]
+        views = [self._make_cand(v, alias, suggestion) for v in views]
         return self.find_matches(word_before_cursor, views, meta='view')
 
     def get_alias_matches(self, suggestion, word_before_cursor):
@@ -616,7 +652,14 @@ class PGCompleter(Completer):
         return self.find_matches(word_before_cursor, self.databases,
                                  meta='database')
 
-    def get_keyword_matches(self, _, word_before_cursor):
+    def get_keyword_matches(self, suggestion, word_before_cursor):
+        keywords = self.keywords_tree.keys()
+        # Get well known following keywords for the last token. If any, narrow
+        # candidates to this list.
+        next_keywords = self.keywords_tree.get(suggestion.last_token, [])
+        if next_keywords:
+            keywords = next_keywords
+
         casing = self.keyword_casing
         if casing == 'auto':
             if word_before_cursor and word_before_cursor[-1].islower():
@@ -625,9 +668,9 @@ class PGCompleter(Completer):
                 casing = 'upper'
 
         if casing == 'upper':
-            keywords = [k.upper() for k in self.keywords]
+            keywords = [k.upper() for k in keywords]
         else:
-            keywords = [k.lower() for k in self.keywords]
+            keywords = [k.lower() for k in keywords]
 
         return self.find_matches(word_before_cursor, keywords,
                                  mode='strict', meta='keyword')
@@ -637,21 +680,21 @@ class PGCompleter(Completer):
         document = Document(text=word_before_cursor,
                             cursor_position=len(word_before_cursor))
         for c in completer.get_completions(document, None):
-            yield Match(completion=c, priority = None)
+            yield Match(completion=c, priority=(0,))
 
     def get_special_matches(self, _, word_before_cursor):
         if not self.pgspecial:
             return []
 
         commands = self.pgspecial.commands
-        cmd_names = commands.keys()
-        desc = [commands[cmd].description for cmd in cmd_names]
-        return self.find_matches(word_before_cursor, cmd_names, mode='strict',
-                                 meta_collection=desc)
+        cmds = commands.keys()
+        cmds = [Candidate(cmd, 0, commands[cmd].description) for cmd in cmds]
+        return self.find_matches(word_before_cursor, cmds, mode='strict')
 
     def get_datatype_matches(self, suggestion, word_before_cursor):
         # suggest custom datatypes
         types = self.populate_schema_objects(suggestion.schema, 'datatypes')
+        types = [self._make_cand(t, False, suggestion) for t in types]
         matches = self.find_matches(word_before_cursor, types, meta='datatype')
 
         if not suggestion.schema:
@@ -689,7 +732,7 @@ class PGCompleter(Completer):
         :param local_tbls: tuple(TableMetadata)
         :return: {TableReference:{colname:ColumnMetaData}}
         """
-
+        ctes = dict((normalize_ref(t.name), t.columns) for t in local_tbls)
         columns = OrderedDict()
         meta = self.dbmetadata
 
@@ -700,6 +743,11 @@ class PGCompleter(Completer):
             columns[tbl].extend(cols)
 
         for tbl in scoped_tbls:
+            # Local tables should shadow database tables
+            if tbl.schema is None and normalize_ref(tbl.name) in ctes:
+                cols = ctes[normalize_ref(tbl.name)]
+                addcols(None, tbl.name, 'CTE', tbl.alias, cols)
+                continue
             schemas = [tbl.schema] if tbl.schema else self.search_path
             for schema in schemas:
                 relname = self.escape_name(tbl.name)
@@ -719,29 +767,36 @@ class PGCompleter(Completer):
                             cols = cols.values()
                             addcols(schema, relname, tbl.alias, reltype, cols)
                             break
-                            
-        # Local tables should shadow database tables
-        for tbl in local_tbls:
-            columns[tbl.name] = tbl.columns
 
         return columns
 
-    def populate_schema_objects(self, schema, obj_type):
-        """Returns list of tables or functions for a (optional) schema"""
-
-        metadata = self.dbmetadata[obj_type]
+    def _get_schemas(self, obj_typ, schema):
+        """ Returns a list of schemas from which to suggest objects
+        schema is the schema qualification input by the user (if any)
+        """
+        metadata = self.dbmetadata[obj_typ]
         if schema:
-            try:
-                objects = metadata[self.escape_name(schema)].keys()
-            except KeyError:
-                # schema doesn't exist
-                objects = []
-        else:
-            schemas = self.search_path
-            objects = [obj for schema in schemas
-                           for obj in metadata[schema].keys()]
+            schema = self.escape_name(schema)
+            return [schema] if schema in metadata else []
+        return self.search_path if self.search_path_filter else metadata.keys()
 
-        return [self.case(o) for o in objects]
+    def _maybe_schema(self, schema, parent):
+        return None if parent or schema in self.search_path else schema
+
+    def populate_schema_objects(self, schema, obj_type):
+        """Returns a list of SchemaObjects representing tables, views, funcs
+        schema is the schema qualification input by the user (if any)
+        """
+
+        return [
+            SchemaObject(
+                name=obj,
+                schema=(self._maybe_schema(schema=sch, parent=schema)),
+                function=(obj_type == 'functions')
+            )
+            for sch in self._get_schemas(obj_type, schema)
+            for obj in self.dbmetadata[obj_type][sch].keys()
+        ]
 
     def populate_functions(self, schema, filter_func):
         """Returns a list of function names
@@ -751,24 +806,20 @@ class PGCompleter(Completer):
         kept or discarded
         """
 
-        metadata = self.dbmetadata['functions']
-
         # Because of multiple dispatch, we can have multiple functions
         # with the same name, which is why `for meta in metas` is necessary
         # in the comprehensions below
-        if schema:
-            schema = self.escape_name(schema)
-            try:
-                return [func for (func, metas) in metadata[schema].items()
-                                for meta in metas
-                                    if filter_func(meta)]
-            except KeyError:
-                return []
-        else:
-            return [func for schema in self.search_path
-                            for (func, metas) in metadata[schema].items()
-                                for meta in metas
-                                    if filter_func(meta)]
+        return [
+            SchemaObject(
+                name=func,
+                schema=(self._maybe_schema(schema=sch, parent=schema)),
+                function=True
+            )
+            for sch in self._get_schemas('functions', schema)
+            for (func, metas) in self.dbmetadata['functions'][sch].items()
+            for meta in metas
+            if filter_func(meta)
+        ]
 
 
 
