@@ -2,6 +2,7 @@ import traceback
 import logging
 import psycopg2
 import psycopg2.extras
+import psycopg2.errorcodes
 import psycopg2.extensions as ext
 import sqlparse
 import pgspecial as special
@@ -145,17 +146,18 @@ class PGExecute(object):
         FROM pg_catalog.pg_database d
         ORDER BY 1'''
 
-    def __init__(self, database, user, password, host, port, dsn):
+    def __init__(self, database, user, password, host, port, dsn, **kwargs):
         self.dbname = database
         self.user = user
         self.password = password
         self.host = host
         self.port = port
         self.dsn = dsn
+        self.extra_args = {k: unicode2utf8(v) for k, v in kwargs.items()}
         self.connect()
 
     def connect(self, database=None, user=None, password=None, host=None,
-            port=None, dsn=None):
+                port=None, dsn=None, **kwargs):
 
         db = (database or self.dbname)
         user = (user or self.user)
@@ -163,42 +165,48 @@ class PGExecute(object):
         host = (host or self.host)
         port = (port or self.port)
         dsn = (dsn or self.dsn)
+        kwargs = (kwargs or self.extra_args)
         pid = -1
         if dsn:
             if password:
                 dsn = "{0} password={1}".format(dsn, password)
             conn = psycopg2.connect(dsn=unicode2utf8(dsn))
             cursor = conn.cursor()
-            # When we connect using a DSN, we don't really know what db,
-            # user, etc. we connected to. Let's read it.
-            db = self._select_one(cursor, 'select current_database()')[0]
-            user = self._select_one(cursor, 'select current_user')[0]
-            host = self._select_one(cursor, 'select inet_server_addr()')[0]
-            port = self._select_one(cursor, 'select inet_server_port()')[0]
         else:
             conn = psycopg2.connect(
-                    database=unicode2utf8(db),
-                    user=unicode2utf8(user),
-                    password=unicode2utf8(password),
-                    host=unicode2utf8(host),
-                    port=unicode2utf8(port))
+                database=unicode2utf8(db),
+                user=unicode2utf8(user),
+                password=unicode2utf8(password),
+                host=unicode2utf8(host),
+                port=unicode2utf8(port),
+                **kwargs)
 
             cursor = conn.cursor()
 
-        cursor.execute("SHOW ALL")
-        db_parameters = dict(name_val_desc[:2] for name_val_desc in cursor.fetchall())
-
-        pid = self._select_one(cursor, 'select pg_backend_pid()')[0]
         conn.set_client_encoding('utf8')
         if hasattr(self, 'conn'):
             self.conn.close()
         self.conn = conn
         self.conn.autocommit = True
+
+        if dsn:
+            # When we connect using a DSN, we don't really know what db,
+            # user, etc. we connected to. Let's read it.
+            # Note: moved this after setting autocommit because of #664.
+            db, user, host, port = self._select_one(
+                cursor,
+                'select current_database(), current_user, inet_server_addr(), inet_server_port()')
+
         self.dbname = db
         self.user = user
         self.password = password
         self.host = host
         self.port = port
+
+        cursor.execute("SHOW ALL")
+        db_parameters = dict(name_val_desc[:2] for name_val_desc in cursor.fetchall())
+
+        pid = self._select_one(cursor, 'select pg_backend_pid()')[0]
         self.pid = pid
         self.superuser = db_parameters.get('is_superuser') == '1'
 
@@ -230,6 +238,18 @@ class PGExecute(object):
         else:
             return json_data
 
+
+    def failed_transaction(self):
+        status = self.conn.get_transaction_status()
+        return status == ext.TRANSACTION_STATUS_INERROR
+
+
+    def valid_transaction(self):
+        status = self.conn.get_transaction_status()
+        return (status == ext.TRANSACTION_STATUS_ACTIVE or
+                status == ext.TRANSACTION_STATUS_INTRANS)
+
+
     def run(self, statement, pgspecial=None, exception_formatter=None,
             on_error_resume=False):
         """Execute the sql in the database and return the results.
@@ -280,10 +300,8 @@ class PGExecute(object):
                 _logger.error("sql: %r, error: %r", sql, e)
                 _logger.error("traceback: %r", traceback.format_exc())
 
-                if (isinstance(e, psycopg2.OperationalError)
+                if (self._must_raise(e)
                         or not exception_formatter):
-                    # Always raise operational errors, regardless of on_error
-                    # specification
                     raise
 
                 yield None, None, None, exception_formatter(e), sql, False
@@ -291,6 +309,23 @@ class PGExecute(object):
                 if not on_error_resume:
                     break
 
+    def _must_raise(self, e):
+        """Return true if e is an error that should not be caught in ``run``.
+
+        ``OperationalError``s are raised for errors that are not under the
+        control of the programmer. Usually that means unexpected disconnects,
+        which we shouldn't catch; we handle uncaught errors by prompting the
+        user to reconnect. We *do* want to catch OperationalErrors caused by a
+        lock being unavailable, as reconnecting won't solve that problem.
+
+        :param e: DatabaseError. An exception raised while executing a query.
+
+        :return: Bool. True if ``run`` must raise this exception.
+
+        """
+        return (isinstance(e, psycopg2.OperationalError) and
+                psycopg2.errorcodes.lookup(e.pgcode) != 'LOCK_NOT_AVAILABLE')
+
     def execute_normal_sql(self, split_sql):
         """Returns tuple (title, rows, headers, status)"""
         _logger.debug('Regular sql statement. sql: %r', split_sql)
@@ -621,4 +656,3 @@ class PGExecute(object):
             cur.execute(query)
             for row in cur:
                 yield row[0]
-
