@@ -1,8 +1,17 @@
 from pygments.token import Token
 from prompt_toolkit.enums import DEFAULT_BUFFER
+from prompt_toolkit.key_binding.vi_state import InputMode
 
+def _get_vi_mode(cli):
+    return {
+        InputMode.INSERT: 'I',
+        InputMode.NAVIGATION: 'N',
+        InputMode.REPLACE: 'R',
+        InputMode.INSERT_MULTIPLE: 'M',
+    }[cli.vi_state.input_mode]
 
-def create_toolbar_tokens_func(get_vi_mode_enabled, get_is_refreshing):
+def create_toolbar_tokens_func(get_vi_mode_enabled, get_is_refreshing,
+                               failed_transaction, valid_transaction):
     """
     Return a function that generates the toolbar tokens.
     """
@@ -31,10 +40,16 @@ def create_toolbar_tokens_func(get_vi_mode_enabled, get_is_refreshing):
                 result.append((token,' (Semi-colon [;] will end the line) '))
 
         if get_vi_mode_enabled():
-            result.append((token.On, '[F4] Vi-mode'))
+            result.append((token.On, '[F4] Vi-mode (' + _get_vi_mode(cli) + ')'))
         else:
             result.append((token.On, '[F4] Emacs-mode'))
 
+        if failed_transaction():
+            result.append((token.Transaction.Failed, '     Failed transaction'))
+
+        if valid_transaction():
+            result.append((token.Transaction.Valid, '     Transaction'))
+
         if get_is_refreshing():
             result.append((token, '     Refreshing completions...'))
 
